@@ -1,73 +1,127 @@
+# --------------------------------------------------------------- Imports ---------------------------------------------------------------- #
+
+# System
 from typing import List, Dict, Optional
 import time, json
 
+# Pip
 from selenium_firefox.firefox import Firefox, By, Keys
 from ktimeout import timeout
 from kcu import strings
 
 from bs4 import BeautifulSoup as bs
 
-YT_URL='https://www.youtube.com'
-YT_STUDIO_URL='https://studio.youtube.com'
-YT_UPLOAD_URL='https://www.youtube.com/upload'
+# ---------------------------------------------------------------------------------------------------------------------------------------- #
 
-MAX_TITLE_CHAR_LEN=100
-MAX_DESCRIPTION_CHAR_LEN=5000
-MAX_TAGS_CHAR_LEN=500
+
+
+# --------------------------------------------------------------- Defines ---------------------------------------------------------------- #
+
+YT_URL          = 'https://www.youtube.com'
+YT_STUDIO_URL   = 'https://studio.youtube.com'
+YT_UPLOAD_URL   = 'https://www.youtube.com/upload'
+YT_LOGIN_URL    = 'https://accounts.google.com/signin/v2/identifier?service=youtube'
+
+MAX_TITLE_CHAR_LEN          = 100
+MAX_DESCRIPTION_CHAR_LEN    = 5000
+MAX_TAGS_CHAR_LEN           = 500
+
+LOGIN_INFO_COOKIE_NAME = 'LOGIN_INFO'
+
+# ---------------------------------------------------------------------------------------------------------------------------------------- #
+
+
+
+# ----------------------------------------------------------- class: Youtube ------------------------------------------------------------- #
 
 class Youtube:
+
+    # ------------------------------------------------------------- Init ------------------------------------------------------------- #
+
     def __init__(
         self,
         cookies_folder_path: str,
         extensions_folder_path: str,
+        email: Optional[str] = None,
+        password: Optional[str] = None,
         host: Optional[str] = None,
         port: Optional[int] = None,
+        full_screen: bool = False,
         disable_images: bool = False
     ):
-        self.browser = Firefox(cookies_folder_path, extensions_folder_path, host=host, port=port, disable_images=disable_images)
+        self.browser = Firefox(cookies_folder_path, extensions_folder_path, host=host, port=port, full_screen=full_screen, disable_images=disable_images)
+        self.channel_id = None
 
         try:
-            self.browser.get(YT_URL)
-            time.sleep(1.5)
-
-            if self.browser.has_cookies_for_current_website():
-                self.browser.load_cookies()
-                time.sleep(1.5)
-                self.browser.refresh()
-            else:
-                input('Log in then press enter')
+            if self.browser.login_via_cookies(YT_URL, LOGIN_INFO_COOKIE_NAME) or self.login(email=email, password=password):
+                time.sleep(0.5)
                 self.browser.get(YT_URL)
-                time.sleep(1.5)
+                time.sleep(0.5)
                 self.browser.save_cookies()
-            
-            self.channel_id = self.get_current_channel_id()
+                time.sleep(0.5)
+                self.channel_id = self.get_current_channel_id()
         except Exception as e:
             print(e)
             self.quit()
 
             raise
 
+
+    # ------------------------------------------------------ Public properties ------------------------------------------------------- #
+
+    @property
+    def is_logged_in(self) -> bool:
+        return self.browser.has_cookie(LOGIN_INFO_COOKIE_NAME)
+
+
+    # -------------------------------------------------------- Public methods -------------------------------------------------------- #
+
+    def login(self, email: Optional[str], password: Optional[str]) -> bool:
+        org_url = self.browser.driver.current_url
+        self.browser.get(YT_URL)
+        time.sleep(0.5)
+
+        logged_in = self.is_logged_in
+
+        if not logged_in and email is not None and password is not None:
+            logged_in = self.__login_auto(email, password)
+        
+        if not logged_in:
+            print('Could not log you in automatically.')
+            logged_in = self.__login_manual()
+
+        if logged_in:
+            time.sleep(0.5)
+            self.browser.get(YT_URL)
+            time.sleep(0.5)
+
+            self.browser.save_cookies()
+
+        time.sleep(0.5)
+        self.browser.get(org_url)
+
+        return logged_in
+
     def upload(
         self,
         video_path: str,
         title: str,
         description: str,
         tags: List[str],
-        _timeout: Optional[int] = 60*3 # 3 min
+        _timeout: Optional[int] = 60*3, # 3 min
+        extra_sleep_after_upload: Optional[int] = None
     ) -> (bool, Optional[str]):
         if _timeout is not None:
             try:
-                return timeout.run(
-                    timeout.partial(self.__upload, video_path, title, description, tags),
-                    _timeout
-                )
+                with timeout.timeout(_timeout):
+                    return self.__upload(video_path, title, description, tags, extra_sleep_after_upload=extra_sleep_after_upload)
             except Exception as e:
                 print(e)
                 self.browser.get(YT_URL)
 
                 return False, None
         else:
-            return self.__upload(video_path, title, description, tags)
+            return self.__upload(video_path, title, description, tags, extra_sleep_after_upload=extra_sleep_after_upload)
 
     def get_current_channel_id(self) -> Optional[str]:
         self.browser.get(YT_URL)
@@ -95,10 +149,8 @@ class Youtube:
     ) -> (bool, bool):
         if _timeout is not None:
             try:
-                return timeout.run(
-                    timeout.partial(self.__comment_on_video, video_id, comment, pinned=pinned),
-                    _timeout
-                )
+                with timeout.timeout(_timeout):
+                    return self.__comment_on_video(video_id, comment, pinned=pinned)
             except Exception as e:
                 print(e)
                 self.browser.get(YT_URL)
@@ -184,12 +236,16 @@ class Youtube:
     def quit(self):
         self.browser.driver.quit()
 
+
+    # ------------------------------------------------------- Private methods -------------------------------------------------------- #
+
     def __upload(
         self,
         video_path: str,
         title: str,
         description: str,
-        tags: List[str]
+        tags: List[str],
+        extra_sleep_after_upload: Optional[int] = None
     ) -> (bool, Optional[str]):
         self.browser.get(YT_URL)
         time.sleep(1.5)
@@ -202,12 +258,9 @@ class Youtube:
             self.browser.find(By.XPATH, "//input[@type='file']").send_keys(video_path)
             print('Upload: uploaded video')
 
-            title_field = self.browser.find(By.ID, 'textbox')
-            title_field.click()
+            title_field = self.browser.find_by('div', id_='textbox')
             time.sleep(0.5)
-            title_field.clear()
-            time.sleep(0.5)
-            title_field.send_keys(Keys.COMMAND + 'a')
+            title_field.send_keys(Keys.BACK_SPACE)
             time.sleep(0.5)
             title_field.send_keys(title[:MAX_TITLE_CHAR_LEN])
             print('Upload: added title')
@@ -253,6 +306,9 @@ class Youtube:
 
             i=0
 
+            if extra_sleep_after_upload is not None and extra_sleep_after_upload > 0:
+                time.sleep(extra_sleep_after_upload)
+
             while True:
                 try:
                     done_button = self.browser.find(By.ID, 'done-button')
@@ -284,14 +340,22 @@ class Youtube:
     # returns (commented_successfully, pinned_comment_successfully)
     def __comment_on_video(self, video_id: str, comment: str, pinned: bool = False) -> (bool, bool):
         self.load_video(video_id)
+        time.sleep(1)
+        self.browser.scroll(150)
 
         try:
-            comments_threads_xpath = "//ytd-comment-thread-renderer[@class='style-scope ytd-item-section-renderer']"
-            self.browser.scroll(350)
+            header = self.browser.find_by('div', id_='masthead-container', class_='style-scope ytd-app')
+            comment_field = self.browser.find(By.XPATH, "//div[@id='placeholder-area']", timeout=5)
 
-            self.browser.find(By.XPATH, "//div[@id='placeholder-area']").click()
-            self.browser.find(By.XPATH, "//div[@id='contenteditable-root']").send_keys(comment)
-            self.browser.find(By.XPATH, "//ytd-button-renderer[@id='submit-button' and @class='style-scope ytd-commentbox style-primary size-default']").click()
+            self.browser.scroll_to_element(comment_field, header_element=header)
+            time.sleep(0.5)
+
+            print('comment: clicking comment_field')
+            comment_field.click()
+            print('comment: sending keys')
+            self.browser.find(By.XPATH, "//div[@id='contenteditable-root']", timeout=0.5).send_keys(comment)
+            print('comment: clicking post_comment')
+            self.browser.find(By.XPATH, "//ytd-button-renderer[@id='submit-button' and @class='style-scope ytd-commentbox style-primary size-default']", timeout=0.5).click()
 
             if not pinned:
                 return True, False
@@ -299,8 +363,13 @@ class Youtube:
             try:
                 dropdown_menu_xpath = "//yt-sort-filter-sub-menu-renderer[@class='style-scope ytd-comments-header-renderer']"
                 dropdown_menu = self.browser.find(By.XPATH, dropdown_menu_xpath)
+                self.browser.scroll_to_element(dropdown_menu, header_element=header)
+                time.sleep(0.5)
 
-                self.browser.find(By.XPATH, "//paper-button[@id='label' and @class='dropdown-trigger style-scope yt-dropdown-menu']", element=dropdown_menu, timeout=2.5).click()
+                dropdown_trigger_xpath = "//paper-button[@id='label' and @class='dropdown-trigger style-scope yt-dropdown-menu']"
+
+                print('comment: clicking dropdown_trigger (open)')
+                self.browser.find(By.XPATH, dropdown_trigger_xpath, element=dropdown_menu, timeout=2.5).click()
 
                 try:
                     dropdown_menu = self.browser.find(By.XPATH, dropdown_menu_xpath)
@@ -310,19 +379,22 @@ class Youtube:
 
                     if last_dropdown_element.get_attribute('aria-selected') == 'false':
                         time.sleep(0.25)
+                        print('comment: clicking last_dropdown_element')
                         last_dropdown_element.click()
                     else:
-                        self.browser.find(By.XPATH, "//paper-button[@id='label' and @class='dropdown-trigger style-scope yt-dropdown-menu']", element=dropdown_menu, timeout=2.5).click()
+                        print('comment: clicking dropdown_trigger (close) (did not click last_dropdown_element (did not find it))')
+                        self.browser.find(By.XPATH, dropdown_trigger_xpath, element=dropdown_menu, timeout=2.5).click()
                 except Exception as e:
                     print(e)
-                    self.browser.find(By.XPATH, "//paper-button[@id='label' and @class='dropdown-trigger style-scope yt-dropdown-menu']", element=dropdown_menu, timeout=2.5).click()
+                    print('comment: clicking dropdown_trigger (close) (did not click last_dropdown_element (error occured))')
+                    self.browser.find(By.XPATH, dropdown_trigger_xpath, element=dropdown_menu, timeout=2.5).click()
             except Exception as e:
                 print(e)
 
-            self.browser.scroll(150)
-            time.sleep(5)
+            # self.browser.scroll(100)
+            time.sleep(2.5)
 
-            for comment_thread in self.browser.find_all(By.XPATH, comments_threads_xpath):
+            for comment_thread in self.browser.find_all(By.XPATH, "//ytd-comment-thread-renderer[@class='style-scope ytd-item-section-renderer']"):
                 pinned_element = self.browser.find(By.XPATH, "//yt-icon[@class='style-scope ytd-pinned-comment-badge-renderer']", element=comment_thread, timeout=0.5)
                 pinned = pinned_element is not None and pinned_element.is_displayed()
 
@@ -330,22 +402,26 @@ class Youtube:
                     continue
 
                 try:
-                    from selenium.webdriver.common.action_chains import ActionChains
-
                     # button_3_dots
                     button_3_dots = self.browser.find(By.XPATH, "//yt-icon-button[@id='button' and @class='dropdown-trigger style-scope ytd-menu-renderer']", element=comment_thread, timeout=2.5)
-                    ActionChains(self.browser.driver).move_to_element(button_3_dots).perform()
+
+                    self.browser.scroll_to_element(button_3_dots, header_element=header)
+                    time.sleep(0.5)
+                    print('comment: clicking button_3_dots')
                     button_3_dots.click()
 
-                    popup_renderer_3_dots = self.browser.find(By.XPATH, "//ytd-menu-popup-renderer[@class='style-scope ytd-popup-container']", timeout=5)
+                    popup_renderer_3_dots = self.browser.find(By.XPATH, "//ytd-menu-popup-renderer[@class='style-scope ytd-popup-container']", timeout=2)
 
                     # dropdown menu item (first)
+                    print('comment: clicking button_3_dots-dropdown_menu_item (to pin it)')
                     self.browser.find(By.XPATH, "//a[@class='yt-simple-endpoint style-scope ytd-menu-navigation-item-renderer']", element=popup_renderer_3_dots, timeout=2.5).click()
 
                     confirm_button_container = self.browser.find(By.XPATH, "//yt-button-renderer[@id='confirm-button' and @class='style-scope yt-confirm-dialog-renderer style-primary size-default']", timeout=5)
 
                     # confirm button
+                    print('comment: clicking confirm_button')
                     self.browser.find(By.XPATH, "//a[@class='yt-simple-endpoint style-scope yt-button-renderer']", element=confirm_button_container, timeout=2.5).click()
+                    time.sleep(2)
 
                     return True, True
                 except Exception as e:
@@ -361,8 +437,57 @@ class Youtube:
 
             return False, False
 
+    def __login_auto(self, email: str, password: str) -> bool:
+        self.browser.get(YT_LOGIN_URL)
+        time.sleep(0.5)
+
+        try:
+            with timeout.timeout(30):
+                email_field = self.browser.find_by('input', { 'id': 'identifierId', 'type': 'email' })
+                email_field.click()
+                self.browser.send_keys_delay_random(email_field, email)
+                time.sleep(1)
+                self.browser.find_by('div', {'jscontroller': 'VXdfxd', 'role': 'button'}).click()
+                time.sleep(1)
+
+                from selenium.webdriver.common.action_chains import ActionChains
+                action = ActionChains(self.browser.driver)
+
+                pass_container = self.browser.find_by('div', { 'id': 'password', 'jscontroller': 'pxq3x' })
+                pass_field = self.browser.find_by('input', { 'type': 'password' }, in_element=pass_container)
+                action.move_to_element(pass_field).perform()
+                pass_field.click()
+                self.browser.send_keys_delay_random(pass_field, password)
+                time.sleep(1)
+                self.browser.find_by('div', {'jscontroller': 'VXdfxd', 'role': 'button'}).click()
+                time.sleep(1)
+        except Exception as e:
+            print('__login_auto', e)
+
+        self.browser.get(YT_URL)
+        time.sleep(0.5)
+
+        return self.is_logged_in
+    
+    def __login_manual(self) -> bool:
+        self.browser.get(YT_LOGIN_URL)
+        time.sleep(0.5)
+
+        try:
+            input('Log in then press return')
+        except Exception as e:
+            print('__login_manual', e)
+
+        self.browser.get(YT_URL)
+        time.sleep(0.5)
+
+        return self.is_logged_in
+    
     def __video_url(self, video_id: str) -> str:
         return YT_URL + '/watch?v=' + video_id
     
     def __channel_videos_url(self, channel_id: str) -> str:
-        return YT_URL + '/channel/' + channel_id + '/videos?view=0&sort=da&flow=grid'
\ No newline at end of file
+        return YT_URL + '/channel/' + channel_id + '/videos?view=0&sort=da&flow=grid'
+
+
+# ---------------------------------------------------------------------------------------------------------------------------------------- #
\ No newline at end of file
